{ File   : VGS2BMP.dpr
  Author : Deniz Oezmen
  Created: 2005-03-12, based on VGSVIEW.PAS created ~2001
  Changed: 2005-10-08

  The Lost Files of Sherlock Holmes: The Case of The Serrated Scalpel VGS
  converter.
}
program VGS2BMP;

{$APPTYPE CONSOLE}

uses
  SysUtils, StrUtils, SaveBMP2;

type
  TSpriteHeader = record
                    Width,
                    Height    : Word;
                    Unknown1  : Byte;
                    Compressed: Boolean;
                    Unknown2,
                    Unknown3  : Byte;
                  end;

  TVGAHeader    = record
                    Unknown1,
                    Unknown2: Longint;
                    VGAID   : array[0..11] of Char
                  end;

  TBuffer       = array of Byte;

var
  VGSFile : file;

function IntToStrL(i: Longint; Len: Byte): string;
begin
  Result := IntToStr(i);

  while Length(Result) < Len do
    Result := '0' + Result
end;

function Prefix(FileName: string): string;
var
  i: Longint;
begin
  FileName := ExpandFileName(FileName);

  i := Length(FileName) + 1;
  repeat
    Dec(i)
  until (i = 0) or (PosEx('.', FileName, i) > 0);

  if (i = 0) or (PosEx('\', FileName, i + 1) > 0) then
    Result := FileName
  else
    Result := Copy(FileName, 1, i - 1)
end;

function OpenVGSFile(var VGSFile: file; VGSName: string): Boolean;
begin
  FileMode := fmShareDenyWrite;
  AssignFile(VGSFile, VGSName);

  {$I-}
  Reset(VGSFile, 1);
  {$I+}
  OpenVGSFile := (IOResult = 0)
end;

function ReadHeader: Boolean;
begin
  Result := True
end;

procedure ConvertPalette3(var Palette3: TPalette3);
var
  i: Byte;
begin
  for i := 0 to 255 do
    with Palette3[i] do
    begin
      R := R shl 2;
      G := G shl 2;
      B := B shl 2
    end
end;

procedure ReadPalette(var Palette3: TPalette3);
begin
  Writeln('Reading palette');

  FillChar(Palette3, SizeOf(Palette3), 0);
  BlockRead(VGSFile, Palette3, SizeOf(Palette3));

  ConvertPalette3(Palette3)
end;

procedure Decode(InBuffer, OutBuffer: TBuffer; InSize, OutSize: Longint;
  CompByte: Byte);
var
  InPos,
  OutPos,
  i     : Longint;
  b, c  : Byte;
begin
  InPos := 0;
  OutPos := 0;

  while InPos < InSize - 1 do
  begin
    b := InBuffer[InPos];
    Inc(InPos);

    if b = CompByte then
    begin
      c := InBuffer[InPos];
      Inc(InPos);
      b := InBuffer[InPos];
      Inc(InPos);

      for i := 0 to b - 1 do
      begin
        OutBuffer[OutPos] := c;
        if OutPos = OutSize - 1 then
          Break;
        Inc(OutPos)
      end;
    end
    else
    begin
      OutBuffer[OutPos] := b;
      if OutPos = OutSize - 1 then
        Break;
      Inc(OutPos)
    end
  end
end;

procedure ReadSprites(VGSName: string; Palette3: TPalette3);
var
  SpriteHeader: TSpriteHeader;
  BMPName     : string;
  InSize      : Word;
  OutSize,
  i           : Longint;
  InBuffer,
  OutBuffer   : TBuffer;
  CompByte,
  SuffixLen   : Byte;
begin
  SuffixLen := 4;
  i := 0;

  while not Eof(VGSFile) do
  begin
    BlockRead(VGSFile, SpriteHeader, SizeOf(SpriteHeader));

    Writeln('  Converting frame #', IntToStrL(i, SuffixLen));

    with SpriteHeader do
    begin
      OutSize := (Width + 1) * (Height + 1);

      if OutSize > 0 then
      begin
        SetLength(OutBuffer, OutSize);

        if Compressed then
        begin
          BlockRead(VGSFile, InSize, 2);
          BlockRead(VGSFile, CompByte, 1);

          Dec(InSize, SizeOf(SpriteHeader) + 3);

          SetLength(InBuffer, InSize);
          BlockRead(VGSFile, InBuffer[0], InSize);
          Decode(InBuffer, OutBuffer, InSize, OutSize, CompByte);

          Finalize(InBuffer)
        end
        else
          BlockRead(VGSFile, OutBuffer[0], OutSize);

        BMPName := Prefix(VGSName) + '#' + IntToStrL(i, SuffixLen) + '.bmp';

        Dump8BitBMP(BMPName, Width + 1, Height + 1, Palette3, OutBuffer);

        Finalize(OutBuffer)
      end
      else
        Writeln('    Empty sprite')
    end;

    Inc(i)
  end
end;

procedure ProcessVGS(VGSName: string);
var
  VGAHeader: TVGAHeader;
  Palette3 : TPalette3;
  i,
  OldPos   : Longint;
begin
  Writeln('Analyzing ' + VGSName);

  // make default palette 
  for i := 0 to 255 do
  begin
    Palette3[i].B := i;
    Palette3[i].G := i;
    Palette3[i].R := i
  end;

  // check for VGA palette 
  OldPos := FilePos(VGSFile);
  {$I-}
  BlockRead(VGSFile, VGAHeader, SizeOf(VGAHeader));
  {$I+}
  if (IOResult <> 0) or (VGAHeader.VGAID <> 'VGA palette' + #026)
  then
    Seek(VGSFile, OldPos)
  else
    ReadPalette(Palette3);

  ReadSprites(VGSName, Palette3)
end;

procedure ParseFiles(FilePara: string);
var
  DirPos,
  i        : Byte;
  FileName,
  Dir,
  VGSName  : string;
  SearchRec: TSearchRec;
  PFilePara: PChar;
begin
  GetMem(PFilePara, Length(FilePara) + 1);
  StrPCopy(PFilePara, FilePara);

  // get the directory information from FilePara 
  DirPos := 0;
  for i := 1 to Length(FilePara) do
    if FilePara[i] in [':', '\'] then
      DirPos := i;

  Dir := Copy(FilePara, 1, DirPos);

  // search for all files matching FilePara 
  if FindFirst(PFilePara, faAnyFile - faDirectory - faVolumeID,
    SearchRec) = 0 then
  begin
    repeat
      FileName := SearchRec.Name;
      VGSName := Dir + FileName;

      // if this file is accessible, analyze it
      if OpenVGSFile(VGSFile, VGSName) then
        if ReadHeader then
        begin
          ProcessVGS(VGSName);
          Close(VGSFile)
        end
        else
          Writeln(VGSName, ' is not a valid VGS file.')
      else
        Writeln(VGSName, ' not accessible')
    until FindNext(SearchRec) <> 0
  end
  else
    Writeln(FilePara, ' not found');
  FreeMem(PFilePara, Length(FilePara) + 1)
end;

begin
  if ParamCount = 1 then
    ParseFiles(ParamStr(1))
  else
    Writeln('Incorrect number of parameters.')
end.
